/*
 * Copyright (c) 1997-2005 Erez Zadok <ezk@cs.stonybrook.edu>
 * Copyright (c) 2001-2005 Stony Brook University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from
 * ftp://ftp.filesystems.org/pub/fistgen/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * File: fistgen/templates/Linux-2.4/inode.c
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_gzipfs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "gzipfs.h"


STATIC int
gzipfs_create(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry);
    PASSERT(hidden_dentry);
    fist_checkinode(dir, "gzipfs_create");

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
    /* XXX this could potentially return a negative hidden_dentry! */
    if (err)
	goto out_lock;

    err = gzipfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

    err = gzipfs_idx_create(hidden_dir_dentry->d_inode, dentry, mode);
    if (err) {
	/*
	 * PROBLEM: we created an empty data file at this point.
	 * should we unlink it?  maybe.  Another solution might be
	 * to change SCA semantics so that a zero-length file need not
	 * have an index file.
	 */
	printk("gzipfs_create: couldn't create idx file %s: %d\n",
	       dtopd(dentry)->idx_dentry->d_name.name, err);
	goto out_lock;
    }
    gzipfs_idx_open(dentry, O_WRONLY);

 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    fist_checkinode(dir, "post gzipfs_create");
    print_exit_status(err);
    return err;
}


STATIC dentry_t *
gzipfs_lookup(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry = NULL;
    const char *name;
    vnode_t *this_vnode;
    dentry_t *this_dir;
    unsigned int namelen;

    print_entry_location();
    hidden_dir_dentry = gzipfs_hidden_dentry(dentry->d_parent); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    namelen = dentry->d_name.len;
    fist_checkinode(dir, "gzipfs_lookup");

    this_vnode = dir;
    this_dir = hidden_dir_dentry;

    fist_print_dentry("gzipfs_lookup IN", dentry);
    fist_print_dentry("gzipfs_lookup: dentry->d_parent IN", dentry->d_parent);
    fist_print_dentry("gzipfs_lookup: hidden_dir_dentry IN", hidden_dir_dentry);
    fist_print_inode("gzipfs_lookup: dir IN", dir);

    if (hidden_dir_dentry->d_inode)
	fist_print_inode("gzipfs_lookup: hidden_dir_dentry->d_inode",
			 hidden_dir_dentry->d_inode);

    /* must initialize dentry operations */
    dentry->d_op = &gzipfs_dops;

    ;

    /* increase refcount of base dentry (lookup_one[_len] will decrement) */
    // THIS IS RIGHT! (don't "fix" it)
    // NO THIS IS WRONG IN 2.3.99-pre6. lookup_one[_len] will NOT decrement
    // dget(hidden_dir_dentry);

    /* will allocate a new hidden dentry if needed */
    hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);

    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    ;

    /* update parent directory's atime */
    fist_copy_attr_atime(dir, hidden_dir_dentry->d_inode);
    /* link the upper and lower dentries */
    dtopd(dentry) = (struct gzipfs_dentry_info *) KMALLOC(sizeof(struct gzipfs_dentry_info), GFP_KERNEL);
    if (!dtopd(dentry)) {
	err = -ENOMEM;
	goto out_dput;
    }
    dtohd(dentry) = hidden_dentry;

    /*
     * Lookup index file.  Note that we create index files only
     * for real files, but at this point we may not know the type
     * of the to-be created inode.  We release idx_dentry's in d_release.
     */
    if (hidden_dentry->d_inode == NULL ||
	S_ISREG(hidden_dentry->d_inode->i_mode)) {
	dtopd(dentry)->idx_dentry = gzipfs_idx_lookup(hidden_dentry);
	if (IS_ERR(dtopd(dentry)->idx_dentry)) {
	    err = PTR_ERR(dtopd(dentry)->idx_dentry);
	    goto out_free;
	}
	/* index file exists and data file does not, or viceversa */
	if ((dtopd(dentry)->idx_dentry->d_inode == NULL) ^
	    (hidden_dentry->d_inode == NULL)) {
	    err = -ENOENT;
	    goto out_free;
	}
    }
    else
	/* no need for index file */
	dtopd(dentry)->idx_dentry = NULL;

    /* lookup is special: it needs to handle negative dentries */
    if (!hidden_dentry->d_inode) {
	d_add(dentry, NULL);
	fist_print_dentry("gzipfs_lookup OUT hidden_dentry", hidden_dentry);
	goto out;
    }

    fist_dprint(6, "lookup \"%s\" -> inode %lu\n", name, hidden_dentry->d_inode->i_ino);
    err = gzipfs_interpose(hidden_dentry, dentry, dir->i_sb, 1);
    if (err)
	goto out_free;

    fist_checkinode(dentry->d_inode, "gzipfs_lookup OUT: dentry->d_inode:");
    fist_checkinode(dir, "gzipfs_lookup OUT: dir");

    fist_print_dentry("gzipfs_lookup OUT hidden_dentry", hidden_dentry);
    fist_print_inode("gzipfs_lookup OUT hidden_inode", hidden_dentry->d_inode);

    if (S_ISREG(hidden_dentry->d_inode->i_mode)) {
	err = gzipfs_idx_open(dentry, O_RDONLY);
	if (err)
	    goto out_free;
	err = gzipfs_idx_read(dentry->d_inode);
	gzipfs_idx_release(dentry->d_inode);
	if (err)
	    goto out_free;
	gzipfs_idx_open(dentry, O_WRONLY);
    }

    /* All is well */
    goto out;

 out_free:
    d_drop(dentry);		/* so that our bad dentry will get destroyed */
    if (dtopd(dentry)->idx_dentry)
	dput(dtopd(dentry)->idx_dentry);
    KFREE(dtopd(dentry));
    dtopd(dentry) = NULL;	/* be safe */

 out_dput:
    if (hidden_dentry)
	dput(hidden_dentry);

 out:
    fist_print_dentry("gzipfs_lookup OUT", dentry);
    print_exit_status(err);
    return ERR_PTR(err);
}


STATIC int
gzipfs_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_old_dentry = gzipfs_hidden_dentry(old_dentry); /* CPW: Moved below print_entry_location */
    hidden_new_dentry = gzipfs_hidden_dentry(new_dentry);

    fist_checkinode(dir, "gzipfs_link-dir");
    fist_checkinode(old_dentry->d_inode, "gzipfs_link-oldinode");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_dir_dentry = lock_parent(hidden_new_dentry);

    /* Also link the index file */
    if (old_dentry->d_inode && S_ISREG(old_dentry->d_inode->i_mode)) {
	/* These must be in the same directory as the main file */
	dentry_t *idx_old_dentry = dtopd(old_dentry)->idx_dentry;
	dentry_t *idx_new_dentry = dtopd(new_dentry)->idx_dentry;

	err = vfs_link(idx_old_dentry,
		       hidden_dir_dentry->d_inode,
		       idx_new_dentry);
	/* XXX this could potentially return a negative dentry! */
	if (err)
	    goto out_lock;

	fist_copy_attr_timesizes(dir, idx_new_dentry->d_inode);
    }

    err = vfs_link(hidden_old_dentry,
		   hidden_dir_dentry->d_inode,
		   hidden_new_dentry);
    if (err || !hidden_new_dentry->d_inode)
	goto out_lock;

    err = gzipfs_interpose(hidden_new_dentry, new_dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    /* propagate number of hard-links */
    old_dentry->d_inode->i_nlink = itohi(old_dentry->d_inode)->i_nlink;

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);
    if (!new_dentry->d_inode)
	d_drop(new_dentry);

    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_unlink(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    inode_t *hidden_dir;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    hidden_dir = itohi(dir); /* CPW: Moved below print_entry_location */
    hidden_dentry = gzipfs_hidden_dentry(dentry);

    PASSERT(hidden_dentry);
    fist_checkinode(dir, "gzipfs_unlink-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    /* first unlink the index file */
    if (dentry->d_inode) {
	if (S_ISREG(dentry->d_inode->i_mode)) {
	    dentry_t *idx_dentry = dtopd(dentry)->idx_dentry;

	    dget(idx_dentry);
	    err = vfs_unlink(hidden_dir, idx_dentry);
	    dput(idx_dentry);
	    if (err)
		printk("SCA unlink \"%s\" failed with error %d\n", dentry->d_name.name, err);
	    fist_print_dentry("gzipfs_unlink: idx_dentry", idx_dentry);
	    if (idx_dentry->d_inode)
		fist_print_inode("gzipfs_unlink: idx_inode", idx_dentry->d_inode);
	}
    } else {
	printk("No index file for %s!\n", dentry->d_name.name);
    }
 out_index:

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_unlink(hidden_dir, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_unlink does that */
	d_delete(hidden_dentry);

 out_lock:
    fist_copy_attr_times(dir, hidden_dir);
    /* propagate number of hard-links */
    dentry->d_inode->i_nlink = itohi(dentry->d_inode)->i_nlink;
    fist_copy_attr_ctime(dentry->d_inode, dir);

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err) {
	//	fist_print_dentry("gzipfs_unlink: 0", dentry);
	d_drop(dentry);
	//	fist_print_dentry("gzipfs_unlink: 1", dentry);
    }

    dput(dentry);

    fist_checkinode(dir, "post gzipfs_unlink-dir");
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "gzipfs_symlink-dir");

    dget(hidden_dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_symlink(hidden_dir_dentry->d_inode,
		      hidden_dentry,
		      symname);
    if (err || !hidden_dentry->d_inode)
	goto out_lock;
    err = gzipfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    fist_checkinode(dir, "post gzipfs_symlink-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "gzipfs_mkdir-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mkdir(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = gzipfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    /* update number of links on parent directory */
    dir->i_nlink = hidden_dir_dentry->d_inode->i_nlink;

    fist_checkinode(dir, "post gzipfs_mkdir-dir");

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_rmdir(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "gzipfs_rmdir-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_rmdir does that */
	d_delete(hidden_dentry);

 out_lock:
    fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
    /* copy the nlink count for our dentry and our parent's dentry */
    dir->i_nlink =  hidden_dir_dentry->d_inode->i_nlink;

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err)
	d_drop(dentry);

    dput(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "gzipfs_mknod-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mknod(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode,
		    dev);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = gzipfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;
    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    fist_checkinode(dir, "post gzipfs_mknod-dir");
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_rename(inode_t *old_dir, dentry_t *old_dentry,
	      inode_t *new_dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_old_dir_dentry;
    dentry_t *hidden_new_dir_dentry;

    print_entry_location();

    hidden_old_dentry = gzipfs_hidden_dentry(old_dentry);/* CPW: Moved below print_entry_location */
    hidden_new_dentry = gzipfs_hidden_dentry(new_dentry);

    fist_checkinode(old_dir, "gzipfs_rename-old_dir");
    fist_checkinode(new_dir, "gzipfs_rename-new_dir");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_old_dir_dentry = get_parent(hidden_old_dentry);
    hidden_new_dir_dentry = get_parent(hidden_new_dentry);
    double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);

    /* Also rename the index file */
    if (old_dentry->d_inode && S_ISREG(old_dentry->d_inode->i_mode)) {
	/* These must be in the same directory as the main file */
	dentry_t *idx_old_dentry = dtopd(old_dentry)->idx_dentry;
	dentry_t *idx_new_dentry = dtopd(new_dentry)->idx_dentry;

	err = vfs_rename(hidden_old_dir_dentry->d_inode, idx_old_dentry,
			 hidden_new_dir_dentry->d_inode, idx_new_dentry);
	if (err)
	    goto out_lock;
    }

    err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
		     hidden_new_dir_dentry->d_inode, hidden_new_dentry);
    if (err)
	goto out_lock;

    fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
    if (new_dir != old_dir)
	fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);

 out_lock:
    // double_unlock will dput the new/old parent dentries whose refcnts
    // were incremented via get_parent above.
    double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);

    fist_checkinode(new_dir, "post gzipfs_rename-new_dir");
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
    int err;
    dentry_t *hidden_dentry;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    fist_print_dentry("gzipfs_readlink IN", dentry);

    if (!hidden_dentry->d_inode->i_op ||
	!hidden_dentry->d_inode->i_op->readlink) {
	err = -EINVAL;
	goto out;
    }

    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry,
						 buf,
						 bufsiz);
    if (err > 0)
	fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);

 out:
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_follow_link(dentry_t *dentry, struct nameidata *nd)
{
    char *buf;
    int len = PAGE_SIZE, err;
    mm_segment_t old_fs;

    print_entry_location();
    //    fist_print_dentry("gzipfs_follow_link dentry IN", dentry);

    buf = KMALLOC(len, GFP_KERNEL);
    if (!buf) {
	err = -ENOMEM;
	goto out;
    }

    /* read the symlink, and then we will follow it */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = dentry->d_inode->i_op->readlink(dentry, buf, len);
    set_fs(old_fs);
    if (err < 0)
	goto out_free;

    buf[err] = 0;	// terminate the buffer -- XXX still needed?

    // XXX: FIXME w/ gzipfs_encode_filename()
    /*
     * vfs_follow_link will increase the nd's mnt refcnt
     * we assume that some other VFS code decrements it.
     */
    err = vfs_follow_link(nd, buf);

 out_free:
    KFREE(buf);
 out:
#if 0
    if (err < 0) {
	dput(nd->dentry);
	printk("EZK follow_link() mnt_cnt %d\n", atomic_read(&nd->mnt->mnt_count));
	mntput(nd->mnt);
    }
#endif

    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_truncate_and_enlarge(dentry_t *dentry, loff_t new_size)
{
    dentry_t *hidden_dentry;
    inode_t *hidden_inode;
    inode_t *inode;
    int err = 0;
    page_t *page = NULL;
    int index, new_index;
    unsigned offset, new_offset, real_size;
    struct encoded_page *ep = NULL, *next_ep, **cur_ep;
    struct scafs_header *hdr;
    void *opaque = 0;
    file_t *hidden_file;
    struct iattr attr;
    off_t abs_offset;
    int need_to_call = 1;
    file_t fake_file; /* needed b/c notify change doesn't get a struct file */

    print_entry_location();

    hidden_dentry = dtohd(dentry);/* CPW: Moved below print_entry_location */
    hidden_inode = hidden_dentry->d_inode;
    inode = dentry->d_inode;
    hdr = &itopd(inode)->hdr;

    down(&hidden_inode->i_sem);

    // calculate page index and page offset (offset)
    new_index = new_size >> PAGE_CACHE_SHIFT;
    new_offset = new_size & ~PAGE_CACHE_MASK;
    if (new_size > hdr->real_size) {
	/* enlarge */
	index = hdr->real_size >> PAGE_CACHE_SHIFT;
	offset = hdr->real_size & ~PAGE_CACHE_MASK;
	/* abs_offset is the absolute offset within the hidden_file
	 * where the encoded representation of the current page begins.
	 * Whee... */
	abs_offset = HIDDEN_PAGE_STARTING_OFFSET(index, inode);
    } else {
	/* shrink */
	index = new_index;
	offset = new_offset;
	abs_offset = HIDDEN_PAGE_STARTING_OFFSET(new_index, inode);
	//truncate_inode_pages(inode->i_mapping, new_size);
    }

    real_size = 0;

    if (offset == 0 || IS_HOLE(index, inode))
	goto page_boundary;

    if (index == new_index)
	offset = new_offset;
    else
	offset = PAGE_CACHE_SIZE;

    /*
     * get our page at offset inode->i_size
     * this is either the page within which we're truncating,
     *		if we are shrinking,
     * or the former last page, if we are enlarging
     */
    /*
     * XXX: set up fake file.  A hack b/c address_ops need a struct file
     * but notify_change only gets a dentry.
     */
    memset(&fake_file, 0, sizeof(file_t));
    // XXX: may need to initialize other fields!
    fake_file.f_dentry = dentry;
    page = gzipfs_get1page(&fake_file, index);
    if (IS_ERR(page)) {
	err = PTR_ERR(page);
	goto out_unlock;
    }

    /* re-encode our page from 0 to "offset" within page */
    kmap(page);
    cur_ep = &ep;
    while (need_to_call) {
	*cur_ep = KMALLOC(sizeof(struct encoded_page), GFP_KERNEL);
	(*cur_ep)->data = (char *) __get_free_page(GFP_KERNEL);
#ifdef FIST_FAST_TAILS
	if (CAN_BE_FASTTAIL_PAGE(offset, inode)) {
	    memcpy((*cur_ep)->data, (char *) page_address(page), offset);
	    need_to_call = 0;
	    err = offset;
	    hdr->flags |= SCA_FLAG_FASTTAIL;
	} else
# endif /* FIST_FAST_TAILS */
	{
	    err = gzipfs_encode_buffers((*cur_ep)->data,
					(char *) page_address(page),
					&need_to_call,
					offset,
					inode,
					inode->i_sb,
					&opaque);
	    if (page->index + 1 >= hdr->num_chunks)
		hdr->flags &= ~SCA_FLAG_FASTTAIL;
	}

	cur_ep = &(*cur_ep)->next;
	*cur_ep = NULL;
	if (err < 0)
	    goto out_free;
	real_size += err;
    }
    gzipfs_idx_set_entry(hdr, index, abs_offset + real_size);
    // write new bytes
    dget(hidden_dentry);
    /*
     * dentry_open will decrement mnt refcnt if err.
     * otherwise fput() will do an mntput() for us upon file close.
     */
    mntget(stopd(inode->i_sb)->hidden_mnt);
    hidden_file = dentry_open(hidden_dentry, stopd(inode->i_sb)->hidden_mnt, O_WRONLY);
    if (IS_ERR(hidden_file)) {
	err = PTR_ERR(hidden_file);
	goto out_free;
    }
    err = gzipfs_copy_encoded_data(hidden_file,
				   abs_offset,
				   ep,
				   real_size);
    fput(hidden_file);
    if (err < 0)
	goto out_free;

 page_boundary:
    index = new_index;
    offset = new_offset;
    abs_offset = HIDDEN_PAGE_STARTING_OFFSET(new_index, inode);

    if (index >= hdr->num_chunks)
	real_size = 0;		/* hole, even though imcomplete */

    // call notify change on lower data file w/ new size
    attr.ia_size = abs_offset + real_size;
    attr.ia_valid = ATTR_SIZE;
    err = notify_change(hidden_dentry, &attr);
    if (err < 0)
	goto out_free;

    // update IT for the new encoded bytes
    hdr->real_size = new_size;
    inode->i_size = new_size;
    if (offset == 0) {
	if (index > 0)
	    err = gzipfs_idx_set_entry(hdr, index - 1, attr.ia_size);
	hdr->num_chunks = index;
    } else {
	err = gzipfs_idx_set_entry(hdr, index, attr.ia_size);
	hdr->num_chunks = index + 1;
    }
    if (err < 0)
	goto out_free;
    mark_inode_dirty(inode);	/* so it'll write the index table */

    /* all is ok */
 out_free:
    /* free encoded_pages */
    while (ep) {
	next_ep = ep->next;
	free_page((unsigned long) ep->data);
	KFREE(ep);
	ep = next_ep;
    }
    kunmap(page);
    if (page)
	page_cache_release(page);
 out_unlock:
    up(&hidden_inode->i_sem);
 out:
    print_exit_location();
    return err;
}


/* We call only the lower-level permission so that file systems with ACLs
 * or more complex permission models work correctly.  This results in
 * O(n^2) calls to permission with n stacked file systems, because each
 * vfs_foo function first calls permission for the top layer, which in
 * turn makes n calls, then calls vfs_foo on the next layer, which will
 * make n calls, etc.
 *
 * An alternative would be to punt permission checking entirely to the
 * lower-level file system, but then we would need to check that we aren't
 * performing any operations without appropriate permissions checking.
 */
STATIC int
gzipfs_permission(inode_t *inode, int mask)
{
    inode_t *hidden_inode;
    int err;

    print_entry_location();

    if (mask & MAY_WRITE) {
        /*
         * Nobody gets write access to a read-only fs.
         */
       if (IS_RDONLY(inode) &&
          (S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) || S_ISLNK(inode->i_mode))) {
            err = -EROFS;
            goto out;
       }
    }

    hidden_inode = itohi(inode);
    err = permission(hidden_inode, mask);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_inode_revalidate(dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry);
    hidden_inode = hidden_dentry->d_inode;
    //    fist_print_dentry("gzipfs_inode_revalidate IN", dentry);

    if (hidden_inode->i_op && hidden_inode->i_op->revalidate) {
	err = hidden_inode->i_op->revalidate(hidden_dentry);
	/*
	 * XXX: CPW: I believe that this should happen unconditionally,
	 * whether or not there is a revalidation method, and also
	 * whether or not it fails.  This leads to better consistency
	 * between the upper and lower-level file systems.
	 * EZK: however, if we do that, then we should validate all other
	 * attributes as well.
 	 */
	if (!err) {
	    fist_copy_attr_all(dentry->d_inode, hidden_inode);
	}
    }

    //    fist_print_dentry("gzipfs_inode_revalidate OUT", dentry);
    print_exit_status(err);
    return err;
}


STATIC int
gzipfs_setattr(dentry_t *dentry, struct iattr *ia)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *inode;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = gzipfs_hidden_dentry(dentry);
    inode = dentry->d_inode;
    hidden_inode = itohi(inode);
    fist_checkinode(inode, "gzipfs_setattr");

    if (ia->ia_valid & ATTR_SIZE) {
	vmtruncate(dentry->d_inode, ia->ia_size);
	err = gzipfs_truncate_and_enlarge(dentry, ia->ia_size);
	if (err < 0)
	    goto out;
	ia->ia_valid &= ~ATTR_SIZE; /* b/c we handled it */
    }
    err = notify_change(hidden_dentry, ia);

#if defined(FIST_FILTER_DATA) || defined(FIST_FILTER_SCA)
 out:
#endif /* FIST_FILTER_DATA || FIST_FILTER_SCA */
    /*
     * The lower file system might has changed the attributes, even if
     * notify_change above resulted in an error(!)  so we copy the
     * hidden_inode's attributes (and a few more) to our inode.
     */
    fist_copy_attr_all(inode, hidden_inode);

    fist_checkinode(inode, "post gzipfs_setattr");
    print_exit_status(err);
    return err;
}


#ifdef NOT_USED_YET
STATIC int
gzipfs_getattr(dentry_t *dentry, struct iattr *ia)
{
    return -ENOSYS;
}
#endif /* NOT_USED_YET */



struct inode_operations gzipfs_symlink_iops =
{
    readlink:	gzipfs_readlink,
    follow_link: gzipfs_follow_link,
    permission:	gzipfs_permission,
    revalidate:	gzipfs_inode_revalidate,
    setattr:	gzipfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	gzipfs_getattr,
#endif

};

struct inode_operations gzipfs_dir_iops =
{
    create:	gzipfs_create,
    lookup:	gzipfs_lookup,
    link:	gzipfs_link,
    unlink:	gzipfs_unlink,
    symlink:	gzipfs_symlink,
    mkdir:	gzipfs_mkdir,
    rmdir:	gzipfs_rmdir,
    mknod:	gzipfs_mknod,
    rename:	gzipfs_rename,
    /* no readlink/follow_link for non-symlinks */
    // off because we have setattr
    //    truncate:	gzipfs_truncate,
    permission:	gzipfs_permission,
    revalidate:	gzipfs_inode_revalidate,
    setattr:	gzipfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	gzipfs_getattr,
#endif
};

struct inode_operations gzipfs_main_iops =
{
    permission:	gzipfs_permission,
    revalidate:	gzipfs_inode_revalidate,
    setattr:	gzipfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	gzipfs_getattr,
#endif
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
